{-# LANGUAGE DeriveFunctor     #-}
{-# LANGUAGE FlexibleInstances #-}

module SortImports.Types where

import Data.Void (Void)
import Text.Megaparsec
import Data.List (intercalate)

class View a where
  view :: a -> String

type Parser = Parsec Void String

data Constructors
  = All
  | Explicit [String]
  deriving (Eq, Ord, Show)

data Export
  = Type String (Maybe Constructors)
  | Function String
  | Operator String
  deriving (Eq, Ord, Show)

data Exposure a
  = Exposed a
  | Hidden a
  deriving (Eq, Functor, Show)

data Qualified
  = Qualified_Pre
  | Qualified_Post
  deriving (Eq, Ord, Show)

data Module = Module
  { _qualified :: Maybe Qualified
  , _name      :: String
  , _alias     :: Maybe String
  , _exports   :: Maybe (Exposure [Export])
  } deriving (Eq, Show)

data LineType
  = CodeLine String
  | ModuleLine Module
  deriving (Ord, Eq, Show)

instance Ord Module where
  (Module _ lhs _ _) `compare` (Module _ rhs _ _) = lhs `compare` rhs

instance View Constructors where
  view All           = ".."
  view (Explicit xs) = intercalate ", " xs

instance View Export where
  view (Type s mc) =
    case mc of
      Nothing -> s
      Just cs -> s <> "(" <> view cs <> ")"
  view (Function s) = s
  view (Operator s) = "(" <> s <> ")"

instance View (Exposure [Export]) where
  view (Hidden xs)  = "hiding " <> view (Exposed xs)
  view (Exposed xs) = "(" <> (intercalate ", " . fmap view $ xs) <> ")"

instance View Module where
  view (Module qualified name alias exports) =
    unwords . filter (/= "") $
      [ "import"
      , if qualified == Just Qualified_Pre then "qualified" else ""
      , name
      , if qualified == Just Qualified_Post then "qualified" else ""
      , maybe "" ("as " <>) alias
      , maybe "" view exports
      ]
